/* Written by Chen Yu-Fang and Hong Chih-Duo                              */
/* Copyright (c) 2010  Academia Sinica	                                  */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/*                                                                        */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

package cav2010.automata;

import cav2010.automata.FiniteAutomaton;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

public class FAState implements Comparable<FAState>{
	Hashtable<String,HashSet<FAState>> next;
	Hashtable<String,HashSet<FAState>> pre;
	public int id;
	
	public FAState(int i){
		id=i;
		next=new Hashtable<String,HashSet<FAState>>();
		pre=new Hashtable<String,HashSet<FAState>>();
	}
	public Iterator<String> nextIt(){
		return next.keySet().iterator();
	}
	public Iterator<String> preIt(){	
		return pre.keySet().iterator();
	}
	public String toString(){
		return "S"+id;
	}
	public int getID(){
		return id;
	}
	public Set<FAState> getNext(String a){
		return next.get(a);
	}
	public Set<FAState> getPre(String a){
		return pre.get(a);
	}
	public void addNext(String a, FAState b, FiniteAutomaton auto){
		if(!next.containsKey(a)){
			HashSet<FAState> S = new HashSet<FAState>();
			S.add(b);
			next.put(a,S);
		}else
			next.get(a).add(b);
	}
	/** 
	 * @param s sate
	 * @return if the set of out-going transitions of this state covers that of s 
	 */
	public boolean covers(FAState s)
	{
		Iterator<String> it = s.nextIt();
		while(it.hasNext())
		{
			String ss = it.next();
			if(!next.containsKey(ss))
				return false;
		}
		return true;
	}
	
	public void addPre(String a, FAState n){
		if(pre.get(a)==null){
			pre.put(a, new HashSet<FAState>());
		}
		pre.get(a).add(n);
	}
	public int compareTo(FAState o) {
		return o.getID()-id;
	}
	public boolean equals(Object o) {
		return ((FAState)o).getID()==id;
	}

}
